"read.FCS" <- function(fileName, FCSobj.name="", fcs.type=NULL,                        
                       fcs.byte.size=2,
                       fcs.signed=TRUE,
                       #fcs.endian = "auto", # c("auto","big","little"),
                       use.FCS.shortnames = FALSE,
                       no.names=FALSE,
                       UseS3=FALSE,
                       MY.DEBUG=TRUE) {

    ## 4 parts:
    ## - HEADER,
    ## - TEXT,
    ## - DATA,
    ## - ANALYSIS (optional)
    
    stopifnot(is.character(fileName), length(fileName)==1, fileName!="")
    con <- file(fileName, open="rb")

    offsets <- readHEADER(con,MY.DEBUG=MY.DEBUG)
    txt <- readTEXT(con, offsets, MY.DEBUG=MY.DEBUG)
    mat    <- readDATA(con, offsets, txt,UseS3=UseS3,fileName,use.FCS.shortnames = use.FCS.shortnames,
                       FCSobj.name,no.names=no.names,MY.DEBUG=MY.DEBUG)

    close(con)
    return(mat)

}

"readHEADER" <- function(con,MY.DEBUG=MY.DEBUG) {

    ## Get Header
    ## numbers imply bytes.  This might not work everywhere.
    seek(con,0)
    metadata.Header <- readChar(con,58)
    indices <- list(version = substr(metadata.Header,1,6),
                    TextStart = as.integer(substr(metadata.Header,10,18)),
                    TextEnd = as.integer(substr(metadata.Header,19,26)),
                    TextLength = (as.integer(substr(metadata.Header,19,26)) -
                                  as.integer(substr(metadata.Header,10,18)) + 1),
                    DataStart = as.integer(substr(metadata.Header,27,34)),
                    DataEnd = as.integer(substr(metadata.Header,35,42)),
                    DataLength = (as.integer(substr(metadata.Header,35,42)) -
                                  as.integer(substr(metadata.Header,27,34)) + 1),
                    AnalysisStart = as.integer(substr(metadata.Header,43,50)),
                    AnalysisEnd = as.integer(substr(metadata.Header,51,58)))

    if (MY.DEBUG) {
        cat("Start/Stop:")
        if (length(indices)==9){
            print(indices)
        } else {
            warning("read.FCS: indices is not the correct length of 9")
        }
    }
    
    return(indices)
}

"readTEXT" <- function(con,offsets,MY.DEBUG=MY.DEBUG){
  ## Read the text section:
  seek(con,offsets[["TextStart"]])
  metadata.Text <- readChar(con,offsets[["TextLength"]])
  metadata.Text <- iconv(metadata.Text, "", "latin1", sub="byte")

  
  metadata.delimiter <- substr(metadata.Text,1,1)

  ## removing all end and beginning spaces
  metadata.Text.verbose <- unlist(strsplit(metadata.Text, split=""))
  pos.space<- which(metadata.Text.verbose==" ")
  first <- 1
  last  <- length(metadata.Text.verbose)
  continue.end <- continue.begin <- TRUE

   
  if (!isTRUE(metadata.Text.verbose[last] == metadata.delimiter)){
    ## There is an error in the metadata
    ## There might be some more metadata, but it will be truncated
    last.pos.delim <- max(which(metadata.Text.verbose == metadata.delimiter))
    warning(paste("Metadata Length Discrepancy: TextLength =",
                  offsets[["TextLength"]],
                  "but valid metadata length =", last.pos.delim, sep=" "))
    warning("The metadata will be truncated up to the last delimiter.")

   
   metadata.Text.verbose <- metadata.Text.verbose[1:last.pos.delim]
    last <- last.pos.delim
  }
  
  while(length(pos.space)>=1 & last>=1 &
        continue.begin==TRUE ||
        continue.end==TRUE) {
    if (max(pos.space)==last) {
      pos.space <- pos.space[pos.space!=max(pos.space)]
      last <- last-1
    }
    continue.end <- metadata.Text.verbose[last] != metadata.delimiter
    if (length(pos.space)>0 & last>0) {
      if(min(pos.space)==first){
        pos.space <- pos.space[pos.space!=min(pos.space)]
        first <- first+1
      }
    }
    continue.begin <- metadata.Text.verbose[first] != metadata.delimiter
  }

  ##if (metadata.delimiter %in% c("\\")) {
  ##  metadata.delimiter <- "\\\\"
  ##}
  if (first>last & MY.DEBUG==TRUE) {
    print("There is no metadata.")
    metadata.Text.list <- NULL
  } else {
    ## There is metadata...
    metadata.Text.verbose<- metadata.Text.verbose[first:last]
    metadata.Text <- paste(metadata.Text.verbose, collapse="")
  
    ## This is a list of strings for the parameters
        metadata.Text.list <-strsplit(substr(metadata.Text,2,nchar(metadata.Text)),
                                            split=metadata.delimiter,fixed=TRUE)[[1]]
    
    metadata.list <- as.list(metadata.Text.list[seq(2,length(metadata.Text.list),by=2)])
    names(metadata.list) <- metadata.Text.list[seq(1, length(metadata.Text.list)-1, by=2)]
    ##  The following should read in all the metadata that is in the file ##

    ## if (MY.DEBUG==TRUE) print("Obtaining the metadata.")
    ## metadata.matrix <- matrix(metadata.Text.list[metadata.Text.list!=""],ncol=2, byrow=TRUE)
    ## metadata.names <- metadata.matrix[,1]
    ## metadata.list <- as.list(metadata.matrix[,2])
    
  }
  metadata.list
}


"readDATA" <- function(con, offsets, txt, fileName, FCSobj.name, UseS3,use.FCS.shortnames,no.names, MY.DEBUG=MY.DEBUG){
    
    if (is.na(txt$"$BYTEORD")) {
        stop("There is no $BYTEORD variable in the metadata.")
    }
    
    ##BYTEORD
    byte.string <- txt$"$BYTEORD"
    file.endian <- switch(byte.string,
                          "4,3,2,1" = "big",
                          "2,1" = "big",
                          "1,2" = "little",     
                          "1,2,3,4" = "little",
                          stop(paste("$BYTEORD", byte.string, "not supported")))
    
    data.endian <- ifelse(.Platform$endian==file.endian,
                          .Platform$endian, "swap")
    
    if (MY.DEBUG==TRUE) {
        print(paste("$BYTEORD", txt$"$BYTEORD", sep=" "))
        print(paste("file endian is", file.endian, sep=" "))
        print(paste("platform endian is", .Platform$endian, sep=" "))
        print(paste("FCS object endian is",data.endian, sep=" "))
    }
    
    if (MY.DEBUG==TRUE){
        print("Reading in data")
    }
    
    dattype <- switch(txt$"$DATATYPE",
                      "I" = "integer",
                      "F" = "numeric",
                      stop(paste("$DATATYPE", offsets$"$DATATYPE","not supported")))
    
    
    if (txt$"$MODE" != "L")
      stop(paste("$MODE", txt$"$MODE","not supported"))
    
    ## Get Data
    if (MY.DEBUG==TRUE){
        print("Reading in data")
    }
    
    
    last.pos <- as.numeric(txt$"$TOT") * as.numeric(txt$"$PAR")
    if (MY.DEBUG==TRUE) {
        print("Checking the size and range of data")
    }
    
    nrpar    <- as.integer(txt$"$PAR")
    metadata.ranges <-  as.integer(readFCSgetPar(unlist(txt), paste("$P", 1:nrpar, "R", sep="")))
    metadata.bitwidth <- unique(as.integer(readFCSgetPar(unlist(txt), paste("$P", 1:nrpar, "B", sep=""))))
    metadata.min <- 0
    metadata.max <- max(metadata.ranges)
    metadata.max
   
    
    if(length(metadata.bitwidth)!=1)
      stop("Sorry, I am expecting the bitwidth to be the same for all parameters")
    
    seek(con,offsets[["DataStart"]], origin="start")
    
    byte.size <- metadata.bitwidth/8
    if (!byte.size %in% c(1,2, 4, 8))
      stop(paste("Don't know how to deal with bitwidth",metadata.bitwidth))
    
    ##n.cell.data <- (offsets[["DataEnd"]]-offsets[["DataStart"]]+1)/byte.size
    n.cell.data <- floor(offsets[["DataLength"]])/byte.size
    cell.data <- readBin(con,dattype,
                         signed=FALSE,
                         n=n.cell.data,
                         size=byte.size,
                         endian=data.endian)
    checked <- FALSE
    read.data.again <- FALSE
    
    if ((last.pos < length(cell.data)) & offsets$version=="FCS3.0") {
        ## only in FCS3.0 files there are extra elements read in
        ## only subset the first elements indicated by the metadata.
        cell.data.red <- cell.data[1:last.pos]
        within.range <- (min(cell.data.red) >= metadata.min) & (max(cell.data.red) <= metadata.max)
        if (within.range==TRUE){
            if (MY.DEBUG==TRUE){
                print(paste("WARNING: Size of data [", length(cell.data),
                            "] is greater than the size noted in metadata [",
                            last.pos,"].", sep=""))
                print("WARNING: The following extra elements of the data are omitted when constructing the data.frame:")
                print(cell.data[(last.pos+1):length(cell.data)])
            }
            cell.data <- cell.data.red
            checked <- TRUE
            read.data.again <- FALSE
            mismatch.length <- FALSE
        }
    }
    if ((checked==FALSE) & (offsets$version=="FCS3.0")) {
        within.range <- (abs(min(cell.data)) >= metadata.min) & (max(cell.data) <= metadata.max)
        mismatch.length <- last.pos!=length(cell.data)
        if ((within.range==FALSE) || (mismatch.length==TRUE)){
            read.data.again <- TRUE
            if (MY.DEBUG==TRUE){
                print(paste("Within metadata range?", within.range, sep=" "))
                print(paste("Size of data [", length(cell.data),
                            "]; Size noted in metadata [", last.pos,
                            "]", sep=""))
            }
        } else {
            if ((within.range==TRUE) & (mismatch.length==FALSE)) {
                read.data.again <- FALSE
                if (MY.DEBUG==TRUE) {
                    print("Data passes range and size check")
                }
            }
        }
    }
    
    ## try a different byte size
    if (read.data.again==TRUE) {
        
        possible.signed <- c(1, 0) ## 1==TRUE, 0==FALSE
        possible.bytesize <- c(1,2,4,8)
        possible.bytesize.signed <- as.list(c(paste(possible.bytesize, possible.signed[1],
                                                    sep=" "),
                                              paste(possible.bytesize, possible.signed[2],
                                                    sep=" ")))
        
        i <- 1
        cell.data.a <- cell.data
        within.range.a <- within.range
        removed.data <- NULL
        mismatch.length.a <- mismatch.length
        
        while ((i <= length(possible.bytesize.signed)) & ((mismatch.length.a == TRUE) || (within.range.a==FALSE))) {
            possibility <- as.numeric(unlist(strsplit(possible.bytesize.signed[[i]],
                                                      split=" ")))
            BSIZE <- possibility[1]
            if (possibility[2]==1){
                SIGNED <- TRUE
            } else {
                if (possibility[2]==0) {
                    SIGNED <- FALSE
                }
            }
            seek(con,offsets[["DataStart"]], origin="start")
            cell.data.a <- readBin(con,dattype,
                                   signed=SIGNED,
                                   n=n.cell.data,
                                   size=BSIZE,
                                   endian=data.endian)
            
            if (last.pos!= length(cell.data.a)){
                if ( (offsets$version=="FCS3.0") & (last.pos <length(cell.data.a))){
                    ## we force the size to be equal to the metadata
                    ## for FCS3.0 files we are only checking the range
                    
                    removed.data <- cell.data.a[(last.pos+1):length(cell.data.a)]
                    cell.data.a <- cell.data.a[1:last.pos]
                    mismatch.length.a <- FALSE
                } else if ((offsets$version!="FCS3.0") || (last.pos > length(cell.data.a)) ){
                    mismatch.length.a <- TRUE
        }
            } else {
                if (last.pos==length(cell.data.a)) {
                    mismatch.length.a <- FALSE
                }
            }
            
            within.range.a <- ((min(cell.data.a) >= metadata.min) &
                               (max(cell.data.a) <= metadata.max))
            i <- i+1
        }
        
    if (offsets$version=="FCS3.0"){
        failure.on.rereads <- (within.range.a==FALSE)
    } else {
        if (offsets$version!="FCS3.0") {
            failure.on.rereads <- (mismatch.length.a==TRUE) || (within.range.a==FALSE)
        }
    }
        
        if(failure.on.rereads==TRUE) {
            cat("Error: Exhausted all fcs.byte.size (1,2,4,8) and fcs.signed (TRUE, FALSE)", "\n",
                "possibilities for reading in data.", "\n",
                "There is a discrepancy between the metadata and the data.  ", "\n",
                "Situation of the first data read with default read.FCS arguments:", "\n",
                "   :Range match=", within.range, "\n",
                "   :size of data=", length(cell.data), "\n",
                "   :size noted in metadata=", last.pos, "\n\n")
            stop("Error: Check the input FCS binary file for metadata and data for discrepancies.")
        } else {
            if (failure.on.rereads==FALSE) {
                cell.data <- cell.data.a
                if (MY.DEBUG==TRUE) {
                    print("Re-read the data again; passed range and size check on these parameters:")
                    print(paste("   fcs.byte.size, fcs.signed, fcs.endian: ", BSIZE, ",", SIGNED, ",", data.endian, sep=""))
                    if (offsets$version=="FCS3.0") {
                        print("   NOTE: This file is FCS version 3.0 and only the range NOT size has been checked.")
                        
                        print(paste("WARNING: Size of data [", length(cell.data) +length(removed.data),
                                    "] is greater than the size noted in metadata [",
                                    last.pos,"].", sep=""))
                        print("WARNING: The following extra elements of the data are omitted when constructing the data.frame:")
                        print(removed.data)
                    }
                }
            }
        }
    }
    ## check the ranges... if there are negatives then try a different fcs.signed
    
    if (MY.DEBUG==TRUE) {
        print(paste("Constructing data.frame from N= ",length(cell.data)))
    }
    cell.df <- as.data.frame(t(matrix(cell.data,
                                      ## JYW: There appears to be a mismatch between
                                      ## what the metadata.list says is the total number of
                                      ## data points = 50,000 ($TOT=10,000 x $PAR=5)
                                      ## and what is read in from offsets
                                      ## DataLength=200,000
                                      ncol=as.integer(txt$"$TOT"))))
    
    if (MY.DEBUG==TRUE) print("Get names")
    
    if (!no.names) {
        ## If we don't want to automatically use names, don't do this.
        ## (usually, we will!)
        if (use.FCS.shortnames) {
            fcs.varnames <- readFCSgetPar(unlist(txt), paste("$P", 1:nrpar, "N", sep=""))
            ##sapply(param.dependent.keywords("$P","N"),get.metadata)
            if (sum(is.na(fcs.varnames))==length(fcs.varnames)){
                cat("NOTE: Short names $PnN are missing", "\n")
                cat("Long names $PnS are used instead.", "\n")
                fcs.varnames <- readFCSgetPar(unlist(txt), paste("$P", 1:nrpar, "S", sep=""))
                ##sapply(param.dependent.keywords("$P","S"), get.metadata)
            }
        } else {
            fcs.varnames <- readFCSgetPar(unlist(txt), paste("$P", 1:nrpar, "S", sep=""))
            ##fcs.varnames <- sapply(param.dependent.keywords("$P","S"),get.metadata)
            if (sum(is.na(fcs.varnames))==length(fcs.varnames)){
                cat("NOTE: Long names $PnS are missing.", "\n")
                cat("Short names $PnN are assigned to the dataset instead.", "\n")
                fcs.varnames <- readFCSgetPar(unlist(txt), paste("$P", 1:nrpar, "N", sep=""))
                ##fcs.varnames <- sapply(param.dependent.keywords("$P","N"),get.metadata)
            }
        }
        if (MY.DEBUG==TRUE) {
            print(paste("Assigning ",length(fcs.varnames),
                        " names to dataframe dims",
                        dim(cell.df)[1]," x ",dim(cell.df)[2]))
        }
        colnames(cell.df) <- fcs.varnames
    }
    
    ##
    ## Get Analysis Section
    ##
    
    ## normally, we do not care much about the analysis section.  In
    ## particular, we are more interested in the data and other
    ## components section.  The analysis section contains histograms and
    ## similar from FCS programs.
    
    ## we currently don't know what to do, and probably don't care.
    ##my.file.has.Analysis.Section <-  FALSE
    ##if (my.file.has.Analysis.Section == TRUE) {
    ## do stuff
    ##}
    
    if (MY.DEBUG==TRUE) print("Constructing return value")
    
   

    
    if (UseS3) {
        ## S3 version
        return(structure(list(data=cell.df,metadata=txt),
                         class="FCS"))
    } else {
        ## when we get S4 classes running, then finish off by:
        ##cell.df <- as.matrix(cell.df)
        ##range <- metadata.ranges
        ##ampliPar <- readFCSgetPar(unlist(txt), paste("$P", 1:nrpar, "E", sep=""))
        ##ampli <- do.call("rbind",lapply(ampliPar,function(x) as.integer(unlist(strsplit(x,",")))))
        ##for (i in 1:nrpar){
        ##    if(ampli[i,1] > 0){
        ##        cell.df[,i] <- 10^((cell.df[,i]/(range[i]-1))*ampli[i,1])
        ##    }
        ## }
        
        return(new("FCS",
                   data=as.matrix(cell.df),
                   metadata=new("FCSmetadata",mode=txt[["$MODE"]],
                     size=as.numeric(txt[["$TOT"]]),
                     nparam=as.numeric(txt[["$PAR"]]),
                     shortnames= unlist(readFCSgetPar(unlist(txt), paste("$P", 1:nrpar, "N", sep=""))),
                     longnames=unlist(readFCSgetPar(unlist(txt), paste("$P", 1:nrpar, "S", sep=""))),
                     paramranges=as.numeric(unlist(readFCSgetPar(unlist(txt), paste("$P", 1:nrpar, "R", sep="")))),
                     filename=fileName,
                     objectname=FCSobj.name,original=TRUE,
                     fcsinfo=txt[!(names(txt) %in% c("$MODE", "$TOT", "$PAR",
                       param.dependent.keywords(txt,"$P", "N"),
                       param.dependent.keywords(txt,"$P", "S"),
                       param.dependent.keywords(txt,"$P", "R") ))])))
    }
}

readFCSgetPar <- function(x, pnam) {
    stopifnot(is.character(x), is.character(pnam)) 
    i <- match(pnam, names(x))
    ##if(any(is.na(i)))
    ##  warning(paste("Parameter(s)", pnam, "not contained in", x))
    return(x[i])
}

param.dependent.keywords <- function(x,prefix,suffix) {
    ## returns a vector of strings
    ## the vector has a length of $PAR (the number of parameters)
    ## each string includes the position in the vector
    paste(prefix,1:as.integer(x$"$PAR"),suffix,sep="")
}    
